package com.oracle.truffle.r.nodes.qirinterface;

import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

import com.oracle.truffle.api.source.Source;
import com.oracle.truffle.api.source.SourceSection;
import com.oracle.truffle.r.nodes.RSyntaxNodeVisitor;
import com.oracle.truffle.r.nodes.access.AccessArgumentNode;
import com.oracle.truffle.r.nodes.access.ConstantNode;
import com.oracle.truffle.r.nodes.access.WriteLocalFrameVariableNode;
import com.oracle.truffle.r.nodes.access.variables.ReadVariableNode;
import com.oracle.truffle.r.nodes.control.*;
import com.oracle.truffle.r.nodes.function.FunctionExpressionNode;
import com.oracle.truffle.r.nodes.function.RCallNode;
import com.oracle.truffle.r.nodes.function.RCallSpecialNode;
import com.oracle.truffle.r.nodes.query.*;
import com.oracle.truffle.r.runtime.RRuntime;
import com.oracle.truffle.r.runtime.context.RContext;
import com.oracle.truffle.r.runtime.nodes.RNode;
import com.oracle.truffle.r.runtime.nodes.RSyntaxNode;

import qir.ast.*;
import qir.ast.data.QIRTcons;
import qir.ast.data.QIRTdestr;
import qir.ast.expression.arithmetic.QIRDiv;
import qir.ast.expression.arithmetic.QIRMinus;
import qir.ast.expression.arithmetic.QIRPlus;
import qir.ast.expression.arithmetic.QIRStar;
import qir.ast.expression.logic.QIRAnd;
import qir.ast.expression.logic.QIRNot;
import qir.ast.expression.logic.QIROr;
import qir.ast.expression.relational.QIREqual;
import qir.ast.expression.relational.QIRLowerOrEqual;
import qir.ast.expression.relational.QIRLowerThan;
import qir.ast.operator.*;

/**
 * This visitor translates a {@link RNode} into a {@link QIRNode}.
 */
public final class QIRTranslateVisitor implements RSyntaxNodeVisitor<QIRNode> {
    public static final QIRTranslateVisitor instance = new QIRTranslateVisitor();

    private QIRTranslateVisitor() {
    }

    private final SourceSection dummy = Source.newBuilder("").name("QIR translate").mimeType(RRuntime.R_APP_MIME).build().createUnavailableSection();

    @Override
    public final QIRNode visit(final IfNode ifNode) {
        return new QIRIf(ifNode.getSourceSection(), ifNode.getCondition().asRSyntaxNode().accept(this), ifNode.getThenPart().asRSyntaxNode().accept(this),
                        ifNode.getElsePart().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final WhileNode whileNode) {
        return new QIRTruffleNode(whileNode.getSourceSection(), whileNode.getSourceSection().getCode());
    }

    @Override
    public final QIRNode visit(final ForNode forNode) {
        return new QIRTruffleNode(forNode.getSourceSection(), forNode.getSourceSection().getCode());
    }

    /**
     * Translation of a sequence of statements. Note: This is also where the STRAD-ASSIGN rules are
     * handled.
     */
    @Override
    public final QIRNode visit(final BlockNode block) {
        QIRNode result = null;
        final RNode[] children = block.getSequence();
        final int len = children.length;

        if (len == 0) // Block is empty
            return QIRNull.getInstance();

        result = children[len - 1].asRSyntaxNode().accept(this); // The last statement of the block
        /*
         * In this loop, we wrap the result with the other statements of the block in reverse order
         * in the way described by the STRAD-ASSIGN rules.
         */
        for (int i = len - 2; i > -1; i--) {
            final RNode child = children[i];
            if (child instanceof WriteLocalFrameVariableNode) { // STRAD-ASSIGN-ID
                final RNode value = ((WriteLocalFrameVariableNode) child).getRhs();

                // If the assignment value reads an argument, then we translate to a lambda.
                result = new QIRLambda(dummy, null, new QIRVariable(dummy, (String) ((WriteLocalFrameVariableNode) child).getName(), null), result);
                // Else we apply STRAD-ASSIGN-ID normally
                if (!(value instanceof AccessArgumentNode))
                    result = new QIRApply(dummy, result, value.asRSyntaxNode().accept(this));
            } else if (child instanceof ReplacementDispatchNode && ((ReplacementDispatchNode) child).lhs instanceof ReadVariableNode) {
                final RNode value = ((ReplacementDispatchNode) child).rhs;

                // If the assignment value reads an argument, then we translate to a lambda.
                result = new QIRLambda(dummy, null, new QIRVariable(dummy, ((ReadVariableNode) ((ReplacementDispatchNode) child).lhs).getIdentifier(), null), result);
                // Else we apply STRAD-ASSIGN-ID normally
                if (!(value instanceof AccessArgumentNode))
                    result = new QIRApply(dummy, result, value.asRSyntaxNode().accept(this));
            } else if (child instanceof ReplacementDispatchNode) { // STRAD-ASSIGN-FIELD and
                                                                   // STRAD-ASSIGN-DOT
                // TODO: Take care of STRAD-ASSIGN-DOT
                // TODO: RCallSpecialNode is not necessarily a write
                final ReplacementDispatchNode repl = (ReplacementDispatchNode) child;
                final RCallSpecialNode lhs = (RCallSpecialNode) repl.lhs;
                final ReadVariableNode receiver = (ReadVariableNode) lhs.getSyntaxArguments()[0];
                final ConstantNode eid = (ConstantNode) lhs.getSyntaxArguments()[1];
                final RNode value = repl.rhs;
                result = new QIRApply(dummy, new QIRLambda(dummy, null, new QIRVariable(dummy, receiver.getIdentifier(), null), result),
                                new QIRTcons(dummy, (String) eid.getValue(), value.asRSyntaxNode().accept(this), receiver.accept(this)));
            } else
                throw new UnsupportedOperationException("Cannot translate \"" + block.getSourceSection() + "\" to QIR: untranslatable sequence.");
        }
        return result;
    }

    @Override
    public final QIRNode visit(final ReadVariableNode var) {
        return new QIRVariable(var.getSourceSection(), var.getIdentifier(), null);
    }

    @Override
    public final QIRNode visit(final WriteLocalFrameVariableNode var) {
        // STRAD-ASSIGN-ID should be handled in BlockNode.
        throw new RuntimeException("Error in translation to QIR: should not have visited an assignment.");
    }

    @Override
    public final QIRNode visit(final FunctionExpressionNode fun) {
        return QIRInterface.RToQIRType(fun.getSourceSection(), fun);
    }

    @Override
    public final QIRNode visit(final RCallNode call) {
        final QIRNode fun = call.getFunctionNode().asRSyntaxNode().accept(this);
        final RSyntaxNode[] args = call.getArguments().getArguments();
        final int nbArgs = args.length;
        QIRNode res = fun;

        if (nbArgs == 0) // call is an application with no arguments
            return new QIRApply(dummy, fun, null);
        for (final RSyntaxNode arg : args)
            res = new QIRApply(dummy, res, arg.accept(this));
        return res;
    }

    @Override
    public final QIRNode visit(final RCallSpecialNode dot) {
        final String name = dot.expectedFunction.getName();
        final List<QIRNode> nodes = Arrays.stream(dot.getSyntaxArguments()).map(arg -> ((RSyntaxNode) arg).accept(this)).collect(Collectors.toList());
        switch (name) {
            case "$":
                return new QIRTdestr(dot.getSourceSection(), nodes.get(0), (String) ((ConstantNode) dot.getSyntaxArguments()[1]).getValue());
            case "+":
                return new QIRPlus(dot.getSourceSection(), nodes.get(0), nodes.get(1));
            case "-":
                return new QIRMinus(dot.getSourceSection(), nodes.get(0), nodes.get(1));
            case "*":
                return new QIRStar(dot.getSourceSection(), nodes.get(0), nodes.get(1));
            case "/":
                return new QIRDiv(dot.getSourceSection(), nodes.get(0), nodes.get(1));
            case "==":
                return new QIREqual(dot.getSourceSection(), nodes.get(0), nodes.get(1));
            case "&":
            case "&&":
                return new QIRAnd(dot.getSourceSection(), nodes.get(0), nodes.get(1));
            case "|":
            case "||":
                return new QIROr(dot.getSourceSection(), nodes.get(0), nodes.get(1));
            case "<=":
                return new QIRLowerOrEqual(dot.getSourceSection(), nodes.get(0), nodes.get(1));
            case "<":
                return new QIRLowerThan(dot.getSourceSection(), nodes.get(0), nodes.get(1));
            case ">=":
                return new QIRNot(dot.getSourceSection(), new QIRLowerThan(dot.getSourceSection(), nodes.get(0), nodes.get(1)));
            case ">":
                return new QIRNot(dot.getSourceSection(), new QIRLowerOrEqual(dot.getSourceSection(), nodes.get(0), nodes.get(1)));
            case "!":
                return new QIRNot(dot.getSourceSection(), nodes.get(0));
            default:
                throw new RuntimeException("Unknown call special node: " + name);
        }
    }

    @Override
    public final QIRNode visit(final ReplacementDispatchNode repl) {
        // STRAD-ASSIGN-FIELD and STRAD-ASSIGN-DOT should be handled in BlockNode.
        throw new RuntimeException("Error in translation to QIR: should not have visited an assignment.");
    }

    /**
     * Translation of a subquery.
     */
    @Override
    public final QIRNode visit(final RQIRWrapperNode query) {
        return RContext.queries.get(query.id);
    }

    @Override
    public final QIRNode visit(final RSelectNode select) {
        return new QIRProject(select.getSourceSection(), select.formatter.asRSyntaxNode().accept(this), select.query.asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final RFromNode from) {
        return new QIRScan(from.getSourceSection(), from.getTable().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final RWhereNode where) {
        return new QIRFilter(where.getSourceSection(), where.getFilter().asRSyntaxNode().accept(this), where.getQuery().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final RGroupNode group) {
        return new QIRGroupBy(group.getSourceSection(), group.getGroup().asRSyntaxNode().accept(this), group.getQuery().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final ROrderNode order) {
        return new QIRSortBy(order.getSourceSection(), order.getOrder().asRSyntaxNode().accept(this), order.getIsAscending().asRSyntaxNode().accept(this),
                        order.getQuery().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final RJoinNode join) {
        return new QIRJoin(join.getSourceSection(), join.getFilter().asRSyntaxNode().accept(this), join.getLeft().asRSyntaxNode().accept(this), join.getRight().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final RLeftJoinNode join) {
        return new QIRLeftJoin(join.getSourceSection(), join.getFilter().asRSyntaxNode().accept(this), join.getLeft().asRSyntaxNode().accept(this), join.getRight().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final RRightJoinNode join) {
        return new QIRRightJoin(join.getSourceSection(), join.getFilter().asRSyntaxNode().accept(this), join.getLeft().asRSyntaxNode().accept(this), join.getRight().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final RLimitNode limit) {
        return new QIRLimit(limit.getSourceSection(), limit.getLimit().asRSyntaxNode().accept(this), limit.getQuery().asRSyntaxNode().accept(this));
    }

    @Override
    public final QIRNode visit(final ConstantNode cst) {
        return QIRInterface.RToQIRType(cst.getSourceSection(), cst.getValue());
    }
}